"""
Module to expose more detailed version info for the installed `cydrogen`
"""
version = "0.0.10"
full_version = version
short_version = version.split('-dev')[0]
git_revision = "91b38942daf88b0e09c3077666831e3f767bacb0"
release = '-dev' not in version and '+' not in version

if not release:
    version = full_version


# cython: language_level=3

from cpython.buffer cimport PyBuffer_FillInfo
from libc.stdint cimport uint8_t
from libc.string cimport memcpy

from ._basekey cimport BaseKey
from ._decls cimport kx_keygen, hydro_kx_keypair, hydro_kx_PUBLICKEYBYTES, hydro_kx_SECRETKEYBYTES, hydro_kx_N_PACKET1BYTES
from ._decls cimport kx_n_1, kx_n_2
from ._exceptions cimport KeyExchangeException
from ._secretbox cimport SecretBoxKey

import base64


KP_SIZE = sizeof(hydro_kx_keypair)


cdef class Psk(BaseKey):
    def __repr__(self):
        return f'Psk({repr(str(self))})'

    cdef uint8_t* ptr(self):
        return <uint8_t*>(self.key.ptr)


cdef class SessionPair:
    def __repr__(self):
        return f'SessionPair(rx={repr(self.rx)}, tx={repr(self.tx)})'


cdef class KxPublicKey:
    def __init__(self, kp):
        if kp is None:
            raise ValueError("pk cannot be None")
        if isinstance(kp, SafeMemory):
            if len(kp) != KP_SIZE:
                raise ValueError(f"safemem must be {KP_SIZE} bytes long")
            self.kp = kp
            return
        if isinstance(kp, str):
            kp = base64.standard_b64decode(kp)
        kp = bytes(kp)
        if len(kp) != hydro_kx_PUBLICKEYBYTES:
            raise ValueError(f"{hydro_kx_PUBLICKEYBYTES} bytes required for public key")
        cdef SafeMemory mem = SafeMemory(KP_SIZE)
        cdef hydro_kx_keypair* kp_ptr = <hydro_kx_keypair*>(<void*>(mem.ptr))
        cdef uint8_t* src = kp
        cdef uint8_t* dst = <uint8_t*>(kp_ptr.pk)
        memcpy(dst, src, hydro_kx_PUBLICKEYBYTES)
        mem.mark_readonly()
        self.kp = mem

    cdef uint8_t* ptr(self):
        cdef hydro_kx_keypair* kp_ptr = <hydro_kx_keypair*>(<void*>(self.kp.ptr))
        return <uint8_t*>(kp_ptr.pk)

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        PyBuffer_FillInfo(buffer, self, self.ptr(), hydro_kx_PUBLICKEYBYTES, 1, flags)

    def __str__(self):
        return base64.standard_b64encode(bytes(self)).decode("utf-8")

    def __repr__(self):
        return f'KxPublicKey({repr(str(self))})'


cdef class KxSecretKey:
    # length hydro_kx_SECRETKEYBYTES
    def __init__(self, kp):
        if kp is None:
            raise ValueError("kp cannot be None")
        if isinstance(kp, SafeMemory):
            if len(kp) != KP_SIZE:
                raise ValueError(f"safemem must be {KP_SIZE} bytes long")
            self.kp = kp
            return
        if isinstance(kp, str):
            kp = base64.standard_b64decode(kp)
        kp = bytes(kp)
        if len(kp) != hydro_kx_SECRETKEYBYTES:
            raise ValueError(f"{hydro_kx_SECRETKEYBYTES} bytes required for secret key")
        cdef SafeMemory mem = SafeMemory(KP_SIZE)
        cdef hydro_kx_keypair* kp_ptr = <hydro_kx_keypair*>(<void*>(mem.ptr))
        cdef uint8_t* src = kp
        cdef uint8_t* dst = <uint8_t*>(kp_ptr.sk)
        memcpy(dst, src, hydro_kx_SECRETKEYBYTES)
        mem.mark_readonly()
        self.kp = mem

    cdef uint8_t* ptr(self):
        cdef hydro_kx_keypair* kp_ptr = <hydro_kx_keypair*>(<void*>(self.kp.ptr))
        return <uint8_t*>(kp_ptr.sk)

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        PyBuffer_FillInfo(buffer, self, self.ptr(), hydro_kx_SECRETKEYBYTES, 1, flags)

    def __str__(self):
        return base64.standard_b64encode(bytes(self)).decode("utf-8")

    def __repr__(self):
        return f'KxSecretKey({repr(str(self))})'


cdef class KxPair:
    def __init__(self, kp):
        if kp is None:
            raise ValueError("kp cannot be None")
        if isinstance(kp, SafeMemory):
            # no need to allocate a new SafeMemory object
            if len(kp) != KP_SIZE:
                raise ValueError(f"safemem must be {KP_SIZE} bytes long")
            self.kp = kp
            return
        if isinstance(kp, str):
            kp = base64.standard_b64decode(kp)
        kp = bytes(kp)
        if len(kp) != KP_SIZE:
            raise ValueError(f"{KP_SIZE} bytes required for keypair")
        self.kp = SafeMemory.from_buffer(kp)

    cdef hydro_kx_keypair* ptr(self):
        return <hydro_kx_keypair*>(<void*>(self.kp.ptr))

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        PyBuffer_FillInfo(buffer, self, self.ptr(), sizeof(hydro_kx_keypair), 1, flags)

    def __str__(self):
        return base64.standard_b64encode(bytes(self)).decode("utf-8")

    def __repr__(self):
        return f'KxPair({repr(str(self))})'

    def __bool__(self):
        return bool(self.kp)

    def __len__(self):
        return len(self.kp)

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, KxPair):
            return False
        cdef KxPair o = <KxPair>other
        return self.kp == o.kp

    cpdef public_key(self):
        return KxPublicKey(self.kp)

    cpdef secret_key(self):
        return KxSecretKey(self.kp)

    @classmethod
    def gen(cls):
        return cls(kx_keygen())


cpdef kx_n_gen_session_and_packet(KxPublicKey peer, Psk psk=None):
    # generate the first packet for the key exchange
    if peer is None:
        raise ValueError("peer cannot be None")
    try:
        rx, tx, packet1 = kx_n_1(peer, psk)
    except RuntimeError as ex:
        raise KeyExchangeException("failed to generate first packet") from ex
    cdef SessionPair pair = SessionPair(rx=SecretBoxKey(rx), tx=SecretBoxKey(tx))
    return pair, packet1


cpdef kx_n_gen_session_from_packet(KxPair static_kp, bytes packet1, Psk psk=None):
    if static_kp is None:
        raise ValueError("static key pair cannot be None")
    if packet1 is None:
        raise ValueError("packet1 cannot be None")
    if len(packet1) != hydro_kx_N_PACKET1BYTES:
        raise ValueError(f"Packet1 must be {hydro_kx_N_PACKET1BYTES} bytes long")
    try:
        rx, tx = kx_n_2(packet1, psk, static_kp)
    except RuntimeError as ex:
        raise KeyExchangeException("failed to generate session from packet") from ex
    return SessionPair(rx=SecretBoxKey(rx), tx=SecretBoxKey(tx))



from collections.abc import Buffer
from typing import Protocol, type_check_only

@type_check_only
class Reader(Protocol):
    """
    Protocol for reading bytes from a source.
    """
    def read(self, length: int = -1) -> bytes:
        """
        Read at most `length` bytes from the source.

        Args:
            length (int): The maximum number of bytes to read. If -1, read until EOF.

        Returns:
            bytes: The bytes read from the source.
        """
        ...

@type_check_only
class Writer(Protocol):
    """
    Protocol for writing bytes to a destination.
    """
    def write(self, buf: Buffer) -> int:
        """
        Write bytes to the destination.

        Args:
            buf (Buffer): The bytes to write.

        Returns:
            int: The number of bytes written.
        """
        ...



import pytest

import cydrogen

PAIR_BYTES = b"\x82P\xef\x19f\x18p\x82\xd4WU$q\xd35\xe3]-<P\xb0?w\x19\xfa9\x8f\x0c)A\xd9Q\xa7\x97\xa8\xa9\xcd\x11\xc7\x99\xb3X\xaf\x9a\xccQTNx\x12\xd1a\xc3\xcb\xd5\xb4P\xf2\xe2+1\xf5R3"
PUBLIC_BYTES = b"\x82P\xef\x19f\x18p\x82\xd4WU$q\xd35\xe3]-<P\xb0?w\x19\xfa9\x8f\x0c)A\xd9Q"
PRIVATE_BYTES = b"\xa7\x97\xa8\xa9\xcd\x11\xc7\x99\xb3X\xaf\x9a\xccQTNx\x12\xd1a\xc3\xcb\xd5\xb4P\xf2\xe2+1\xf5R3"


def test_gen_kx_pair():
    pair = cydrogen.KxPair.gen()
    assert bool(pair)
    assert len(pair) == 64


def test_create_kx_pair_from_bytes():
    pair = cydrogen.KxPair(PAIR_BYTES)
    assert bytes(pair) == PAIR_BYTES
    assert bool(pair)
    assert bytes(pair.public_key()) == PUBLIC_BYTES
    assert bytes(pair.secret_key()) == PRIVATE_BYTES


def test_create_zero_key():
    with pytest.raises(TypeError):
        cydrogen.KxPair()
    with pytest.raises(ValueError):
        cydrogen.KxPair(None)
    with pytest.raises(ValueError):
        cydrogen.KxPair(b"")


def test_kx_n():
    pair = cydrogen.KxPair(PAIR_BYTES)
    session, pkt = cydrogen.kx_n_gen_session_and_packet(pair.public_key())
    session_peer = cydrogen.kx_n_gen_session_from_packet(pair, pkt)
    assert session.rx == session_peer.tx
    assert session.tx == session_peer.rx



"""
The `cydrogen` package provides Python bindings for the [libhydrogen](https://github.com/jedisct1/libhydrogen) library.

It includes functionalities for cryptographic operations such as hashing, symmetric
encryption, and signing.

Everything present directly in the parent `cydrogen` module is considered the public API.

Users should not import anything from `cydrogen._someinternal` submodules, as these are
internal implementations and may change without notice.
"""

from ._basekey import BaseKey
from ._context import Context
from ._decls import (
    gen_random_buffer,
    pad,
    random_u32,
    random_uniform,
    randomize_buffer,
    shuffle_buffer,
    unpad,
)
from ._exceptions import (
    CyException,
    DecryptException,
    DeriveException,
    EncryptException,
    KeyExchangeException,
    SignException,
    VerifyException,
)
from ._hash import Hash, HashKey, hash_file
from ._kx_n import KxPair, KxPublicKey, KxSecretKey, Psk, SessionPair, kx_n_gen_session_and_packet, kx_n_gen_session_from_packet
from ._masterkey import MasterKey
from ._secretbox import EncryptedMessage, SecretBox, SecretBoxKey
from ._sign import Signer, SignKeyPair, SignPublicKey, SignSecretKey, Verifier, sign_file, verify_file

__all__ = [
    # random
    "random_u32",
    "random_uniform",
    "randomize_buffer",
    "gen_random_buffer",
    "shuffle_buffer",
    # base key
    "BaseKey",
    # exceptions
    "CyException",
    "EncryptException",
    "DecryptException",
    "DeriveException",
    "SignException",
    "VerifyException",
    "KeyExchangeException",
    # context
    "Context",
    # hashing
    "Hash",
    "HashKey",
    "hash_file",
    # cryptobox
    "SecretBoxKey",
    "SecretBox",
    "EncryptedMessage",
    # master key / derivation
    "MasterKey",
    # signing
    "SignPublicKey",
    "SignSecretKey",
    "SignKeyPair",
    "Signer",
    "Verifier",
    "sign_file",
    "verify_file",
    # key exchange
    "KxPair",
    "KxPublicKey",
    "KxSecretKey",
    "Psk",
    "SessionPair",
    "kx_n_gen_session_and_packet",
    "kx_n_gen_session_from_packet",
    # padding
    "pad",
    "unpad",
]



from collections.abc import Buffer
from dataclasses import dataclass
from typing import Self

from ._basekey import BaseKey
from ._secretbox import SecretBoxKey

class Psk(BaseKey):
    """
    The Psk class represents a pre-shared key (PSK).

    Pre-shared keys are used optionally by the key exchange algorithms.
    """

@dataclass
class SessionPair:
    """
    The SessionPair class represents a pair of symmetric keys to be used to secure a session with a peer.

    Attributes:
        rx: The symmetric key used to decrypt messages received from the peer.
        tx: The symmetric key used to encrypt messages sent to the peer.
    """

    rx: SecretBoxKey
    tx: SecretBoxKey

class KxPublicKey:
    """
    The KxPublicKey class represents a public key used in key exchange.

    KxPublicKey implements the buffer protocol, allowing it to be used as a byte-like object.
    """
    def __init__(self, kp: bytes | str | Buffer | Self):
        """
        Initializes a KxPublicKey instance.

        Args:
            kp: The public key data to initialize the KxPublicKey instance with.

        Raises:
            ValueError: If the provided key data is invalid.
        """
        ...

    def __str__(self) -> str:
        """
        Returns a string representation of the KxPublicKey (base64 encoded).

        Returns:
            A base64 encoded string representation of the public key.
        """
        ...

    def __repr__(self) -> str: ...
    def __buffer__(self, flags: int, /) -> memoryview: ...

class KxSecretKey:
    """
    The KxSecretKey class represents a secret key used in key exchange.

    KxSecretKey implements the buffer protocol, allowing it to be used as a byte-like object.
    """
    def __init__(self, kp: bytes | str | Buffer | Self):
        """
        Initializes a KxSecretKey instance.

        Args:
            kp: The secret key data to initialize the KxSecretKey instance with.

        Raises:
            ValueError: If the provided key data is invalid.
        """
        ...

    def __str__(self) -> str:
        """
        Returns a string representation of the KxSecretKey (base64 encoded).

        Returns:
            A base64 encoded string representation of the secret key.
        """
        ...

    def __repr__(self) -> str: ...
    def __buffer__(self, flags: int, /) -> memoryview: ...

class KxPair:
    """
    KxPair represents a pair of public/secret keys used in key exchange.

    KxPair implements the buffer protocol, allowing it to be used as a byte-like object.
    """
    def __init__(self, kp: str | bytes | Buffer | Self):
        """
        Initializes a KxPair instance.

        Args:
            kp: The key pair data to initialize the KxPair instance with.

        Raises:
            ValueError: If the provided key pair data is invalid.
        """
        ...

    def __str__(self) -> str:
        """
        Returns a string representation of the KxPair (base64 encoded).

        Returns:
            A base64 encoded string representation of the key pair.
        """
        ...

    def public_key(self) -> KxPublicKey:
        """
        Returns the public key part of the KxPair.

        Returns:
            A KxPublicKey instance representing the public key part of the KxPair.
        """
        ...

    def secret_key(self) -> KxSecretKey:
        """
        Returns the secret key part of the KxPair.

        Returns:
            A KxSecretKey instance representing the secret key part of the KxPair.
        """
        ...

    @classmethod
    def gen(cls) -> Self:
        """
        Generates a new KxPair instance with a random key pair.

        Returns:
            A new KxPair instance with a randomly generated public and secret key.
        """
        ...

    def __repr__(self) -> str: ...
    def __buffer__(self, flags: int, /) -> memoryview: ...

def kx_n_gen_session_and_packet(peer: KxPublicKey, psk: Psk | None = None) -> tuple[SessionPair, bytes]:
    """
    Generate session keys and a packet with an ephemeral public key to send to the server

    Args:
        peer: The public key of the peer to exchange keys with.
        psk: An optional pre-shared key to use in the key exchange.

    Returns:
        A tuple containing a SessionPair with the symmetric keys for the session and a bytes object
        representing the packet to send to the server.

    Raises:
        KeyExchangeException: If the operation fails.
    """
    ...

def kx_n_gen_session_from_packet(static_kp: KxPair, packet1: bytes, psk: Psk | None = None) -> SessionPair:
    """
    Generate session keys from a received packet and the static key pair.

    Args:
        static_kp: The static key pair used for the key exchange.
        packet1: The received packet containing the ephemeral public key and other data.
        psk: An optional pre-shared key to use in the key exchange.

    Returns:
        A SessionPair with the symmetric keys for the session.

    Raises:
        KeyExchangeException: If the operation fails. In particular, this exception is raised if the packet is invalid or if the key exchange fails.
    """
    ...



# cython: language_level=3

from libc.stdint cimport uint8_t

from ._basekey cimport BaseKey
from ._decls cimport hydro_kx_keypair
from ._secretbox cimport SecretBoxKey
from ._utils cimport SafeMemory

from dataclasses import dataclass


cdef class Psk(BaseKey):
    cdef uint8_t* ptr(self)


@dataclass
cdef class SessionPair:
    cdef readonly SecretBoxKey rx
    cdef readonly SecretBoxKey tx


cdef class KxPublicKey:
    cdef SafeMemory kp

    cdef uint8_t* ptr(self)


cdef class KxSecretKey:
    cdef SafeMemory kp

    cdef uint8_t* ptr(self)


cdef class KxPair:
    cdef SafeMemory kp

    cdef hydro_kx_keypair* ptr(self)
    cpdef public_key(self)
    cpdef secret_key(self)

cpdef kx_n_gen_session_and_packet(KxPublicKey peer, Psk psk=*)
cpdef kx_n_gen_session_from_packet(KxPair static, bytes packet1, Psk psk=*)



class CyException(Exception):
    """
    Base class for all exceptions in the `cydrogen` package.
    """

    ...

class EncryptException(CyException):
    """
    Exception raised for errors during encryption operations.
    """

    ...

class DecryptException(CyException):
    """
    Exception raised for errors during decryption operations.

    In particular, this exception is raised when the decryption fails due to an invalid key
    or tampered ciphertext.
    """

    ...

class DeriveException(CyException):
    """
    Exception raised for errors during key derivation operations.
    """

    ...

class SignException(CyException):
    """
    Exception raised for errors during signing operations.
    """

    ...

class VerifyException(CyException):
    """
    Exception raised for errors during signature verification operations.

    In particular, this exception is raised when the signature verification fails due to an
    invalid signature or tampered data.
    """

    ...

class KeyExchangeException(CyException):
    """
    Exception raised for errors during key exchange operations.
    """

    ...
